#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains helpful functions for upgrading the user.json file
from previous versions to the current version.

An explination of how we used to do things:
1. We had a `user.json` file, which either didn't have a version number or was version 1
2. We added fields to the `user.json` file when the user loads it, 
   setting them equal to the new default values
3. We never made any backwards incompatible changes

However, we now want to make some backwards incompatible changes. We do that by having mulitple
schemas for the user.json file, and having an upgrade process that upgrades from one schema to 
the other.

1. For new users who show up, they will just get a default `user.json` file. 
   It will have the most up to date version number in it, obviously.

2. When old users show up, we compare the version of the user.json that they 
   have to the current version of the user.json.
   - If they have a user.json with _no version_ number or _version 1_, then we make sure they
     have the "final schema" for version 1. We only need to do this because of how we used to 
     change version 1 without making new versions.
   - Then, if they have an old version of the user.json, then we actually do an upgrade from one 
     version to the next. Simply using an upgrade script that converts.

_Any change the user.json format must be met with a version bump._
"""
from typing import Any, Dict
from mitosheet.experiments.experiment_utils import get_current_experiment, get_new_experiment
from mitosheet._version import package_name

from mitosheet.user.db import get_user_json_object, set_user_json_object
from mitosheet.user.schemas import (UJ_CLOSED_FEEDBACK, UJ_EXPERIMENT, UJ_FEEDBACKS_V2,
                                    UJ_INTENDED_BEHAVIOR,
                                    UJ_MITOSHEET_LAST_FIFTY_USAGES,
                                    UJ_MITOSHEET_LAST_FIVE_USAGES,
                                    UJ_MITOSHEET_PRO, UJ_MITOSHEET_TELEMETRY, UJ_RECEIVED_CHECKLISTS,
                                    UJ_USER_JSON_VERSION, USER_JSON_DEFAULT,
                                    USER_JSON_VERSION_1)


def try_upgrade_to_final_user_json_version_1(user_json_object: Dict[str, Any]) -> Dict[str, Any]:
    """
    With user_json_version = 1, we used to make changes to the user.json 
    without changing the version number. We simply just added a bunch
    of new fields.

    Thus, if a user is on user_json_version = 1, before upgrading the user 
    json to a later version, we first make sure they have all the fields
    that the final user_json_version = 1 had.

    NOTE: this also handles the user.json that is written by the installer, 
    which is effectively just the static_user_id and nothing else.
    """

    # If the user is on an incredibly old version, we might not have a version number
    # in which case we just set it equal to 1, and continue with the upgrade
    if UJ_USER_JSON_VERSION not in user_json_object:
        user_json_object[UJ_USER_JSON_VERSION] = 1
    
    # We only need to upgrade to the final version of 1 in the case
    # that they are actually in version 1!
    if user_json_object[UJ_USER_JSON_VERSION] == 1:
            
        # Then we just make sure that the user.json has all the fields it needs defined
        # in the final version one schema, and if they are not defined, it sets them 
        # to the default values provided
        for field, default_value in USER_JSON_VERSION_1.items():
            if field not in user_json_object:
                user_json_object[field] = default_value

    return user_json_object

def upgrade_final_user_json_version_1_to_2(final_user_json_version_1: Dict[str, Any]) -> Dict[str, Any]:
    """
    Two changes in this version: 
    1. removes the intended_behavior field, as we no longer user it
    2. removes the closed_feedback field, as we no longer user it
    3. renames mitosheet_last_five_usages -> mitosheet_last_fifty_usages, 
       so we can now track 50 usages
    """
    if final_user_json_version_1[UJ_USER_JSON_VERSION] != 1:
        return final_user_json_version_1

    # 0: bump the version number
    final_user_json_version_1[UJ_USER_JSON_VERSION] = 2
    
    # 1: Remove intended behavior field
    del final_user_json_version_1[UJ_INTENDED_BEHAVIOR]

    # 2: Remove closed_feedback field
    del final_user_json_version_1[UJ_CLOSED_FEEDBACK]

    # 3: Rename mitosheet_last_five_usages -> mitosheet_last_fifty_usages
    final_user_json_version_1[UJ_MITOSHEET_LAST_FIFTY_USAGES] = final_user_json_version_1[UJ_MITOSHEET_LAST_FIVE_USAGES]
    del final_user_json_version_1[UJ_MITOSHEET_LAST_FIVE_USAGES]

    return final_user_json_version_1

def upgrade_user_json_version_2_to_3(user_json_version_2: Dict[str, Any]) -> Dict[str, Any]:
    """
    Just adds the user json telemetry field, which may already
    be set (due to it being added to the installer). Thus, only
    if it doesn't exist, we set it to True
    """
    # First, bump the version number
    user_json_version_2[UJ_USER_JSON_VERSION] = 3

    # Then, set new field
    if UJ_MITOSHEET_TELEMETRY not in user_json_version_2:
        user_json_version_2[UJ_MITOSHEET_TELEMETRY] = True

    return user_json_version_2

def upgrade_user_json_version_3_to_4(user_json_version_3: Dict[str, Any]) -> Dict[str, Any]:
    """
    Just adds the user json feedback_v2 field
    """
    # First, bump the version number
    user_json_version_3[UJ_USER_JSON_VERSION] = 4

    # Then, set new field
    user_json_version_3[UJ_FEEDBACKS_V2] = {}
    return user_json_version_3

def upgrade_user_json_version_4_to_5(user_json_version_4: Dict[str, Any]) -> Dict[str, Any]:
    """
    Just adds the UJ_MITOSHEET_PRO field, and sets it to False,
    if it is not already in the user json (which it may be,
    as it is now set by the installer). 
    """
    # First, bump the version number
    user_json_version_4[UJ_USER_JSON_VERSION] = 5

    # Then, set new field
    if UJ_MITOSHEET_PRO not in user_json_version_4:
        user_json_version_4[UJ_MITOSHEET_PRO] = False
    return user_json_version_4

def upgrade_user_json_version_5_to_6(user_json_version_5: Dict[str, Any]) -> Dict[str, Any]:
    """
    Just adds the UJ_EXPERIMENT field, and sets it equal to
    the default expirments list
    """
    # First, bump the version number
    user_json_version_5[UJ_USER_JSON_VERSION] = 6

    # Then, set new field
    if UJ_EXPERIMENT not in user_json_version_5:
        user_json_version_5[UJ_EXPERIMENT] = get_new_experiment()

    return user_json_version_5

def upgrade_user_json_version_6_to_7(user_json_version_6: Dict[str, Any]) -> Dict[str, Any]:
    """
    Just adds the UJ_RECEIVED_CHECKLISTS field, and sets it equal to the full onboarding
    checklist. This is because we do not want users to have to go back and do the onboarding
    checklist if they are already onboarded.

    We only upgrade them if they are not on the installer_communication_and_time_to_value 
    experiment, as this means that they have used the installer, and we actually _want_
    them to go through the onboarding checklist.

    We also count the checklist as completed if the user is on mitosheet-private, as this upgrade
    does not happen through the installer, and so this experiment id will be set, but we do not
    want these users to go through the checklist. This is a temporary fix, and we need a much better
    long-term solution going forward.
    """
    # First, bump the version number
    user_json_version_6[UJ_USER_JSON_VERSION] = 7

    # Then, set new field
    if UJ_RECEIVED_CHECKLISTS not in user_json_version_6:
        current_experiment = user_json_version_6[UJ_EXPERIMENT]
        if current_experiment['experiment_id'] != 'installer_communication_and_time_to_value' or package_name == 'mitosheet-private':
            user_json_version_6[UJ_RECEIVED_CHECKLISTS] = {
                "onboarding_checklist": ['signup', 'import', 'filter', 'pivot', 'graph', 'finalize']
            }
        else:
            user_json_version_6[UJ_RECEIVED_CHECKLISTS] = {}


    return user_json_version_6

def try_upgrade_user_json_to_current_version() -> None:
    user_json_object = get_user_json_object()

    if user_json_object is None:
        user_json_object = USER_JSON_DEFAULT
    
    # If we don't need to upgrade, don't upgrade
    if UJ_USER_JSON_VERSION in user_json_object and user_json_object[UJ_USER_JSON_VERSION] == USER_JSON_DEFAULT[UJ_USER_JSON_VERSION]:
        return

    # Try to upgrade it to the final version of 1, if it is not already there
    try_upgrade_to_final_user_json_version_1(user_json_object)

    # Then, upgrade from version 1 to version 2, which is the 
    # most up-to-date version
    if user_json_object[UJ_USER_JSON_VERSION] == 1:
        user_json_object = upgrade_final_user_json_version_1_to_2(user_json_object)
    if user_json_object[UJ_USER_JSON_VERSION] == 2:
        user_json_object = upgrade_user_json_version_2_to_3(user_json_object)
    if user_json_object[UJ_USER_JSON_VERSION] == 3:
        user_json_object = upgrade_user_json_version_3_to_4(user_json_object)
    if user_json_object[UJ_USER_JSON_VERSION] == 4:
        user_json_object = upgrade_user_json_version_4_to_5(user_json_object)
    if user_json_object[UJ_USER_JSON_VERSION] == 5:
        user_json_object = upgrade_user_json_version_5_to_6(user_json_object)
    if user_json_object[UJ_USER_JSON_VERSION] == 6:
        user_json_object = upgrade_user_json_version_6_to_7(user_json_object)

    # We always make sure that the experiment is the most up to date
    # version of the experiment
    new_experiment = get_new_experiment()
    if new_experiment is not None and user_json_object[UJ_EXPERIMENT]['experiment_id'] != new_experiment['experiment_id']:
            user_json_object[UJ_EXPERIMENT] = new_experiment

    set_user_json_object(user_json_object)






#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import inspect
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

from IPython import get_ipython
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.import_steps.dataframe_import_code_chunk import \
    DataframeImportCodeChunk
from mitosheet.errors import make_dataframe_not_found_error
from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param


def get_variable_with_name_from_caller(variable_name: str) -> Optional[Any]:
    """
    This helper function tries to find a variable with in the calling context. 
    It will first check ipython to see if it is defined, and get the variable
    from this context. 

    If ipython is not defined, it will assume we are in some other python environment,
    and attempt to get the variable from the callers stack frames.

    This is useful so that we can test this function (e.g. pytest does not run in
    an ipython environment).
    """
    ipython = get_ipython() # type: ignore
    if ipython is not None:
        return ipython.ev(variable_name)

    stack = inspect.stack()
    for s in reversed(stack):
        local_var = s[0].f_locals.get(variable_name)
        if local_var is not None:
            return local_var
            
        global_var = s[0].f_globals.get(variable_name)
        if global_var is not None:
            return global_var

    return None

class DataframeImportStepPerformer(StepPerformer):
    """
    Allows you to import a dataframe
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'dataframe_import'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        df_names: str = get_param(params, 'df_names')

        # We make a new state to modify it
        post_state = prev_state.copy()

        pandas_start_time = perf_counter()
        
        # Get the dataframe, and import it

        for df_name in df_names:
            try:
                df = get_variable_with_name_from_caller(df_name)
            except NameError:
                raise make_dataframe_not_found_error(df_name)

            # TODO: There is a bug if you import the same dataframe twice, then you get
            # issues where the generated code does not match with the sheet. Do we want 
            # to insist on uniqueness here? Or do we want to automatically make a copy?
            post_state.add_df_to_state(df, DATAFRAME_SOURCE_IMPORTED, df_name=df_name)
        
        pandas_processing_time = perf_counter() - pandas_start_time


        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'result': {}
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            DataframeImportCodeChunk(
                prev_state, 
                post_state, 
                'Imported Dataframes',
                'Imported dataframes into the mitosheet'
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the The Mito Enterprise license.

from typing import Any, Dict, Optional, Union
import os
from mitosheet.telemetry.telemetry_utils import log
from mitosheet.types import CodeSnippetEnvVars

# Note: Do not change these keys, we need them for looking up 
# the environment variables from previous mito_config_versions.
MITO_CONFIG_VERSION = 'MITO_CONFIG_VERSION'
MITO_CONFIG_SUPPORT_EMAIL = 'MITO_CONFIG_SUPPORT_EMAIL'
MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL = 'MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL' 
MITO_CONFIG_CODE_SNIPPETS_VERSION  = 'MITO_CONFIG_CODE_SNIPPETS_VERSION' 
MITO_CONFIG_CODE_SNIPPETS_URL = 'MITO_CONFIG_CODE_SNIPPETS_URL'
MITO_CONFIG_DISABLE_TOURS = 'MITO_CONFIG_DISABLE_TOURS'

# Note: The below keys can change since they are not set by the user.
MITO_CONFIG_CODE_SNIPPETS = 'MITO_CONFIG_CODE_SNIPPETS'

# The default values to use if the mec does not define them
DEFAULT_MITO_CONFIG_SUPPORT_EMAIL = 'founders@sagacollab.com'
DEFAULT_MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL = 'founders@sagacollab.com'

def upgrade_mec_1_to_2(mec: Dict[str, Any]) -> Dict[str, Any]:
    """
    Converts mec of shape:
    {
        'MITO_CONFIG_VERSION': '1',
        'MITO_CONFIG_SUPPORT_EMAIL': 'support@mito.com'
    }
    
    into: 

    {
        'MITO_CONFIG_VERSION': '2',
        'MITO_CONFIG_SUPPORT_EMAIL': 'support@mito.com',
        'MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL': None,
        'MITO_CONFIG_CODE_SNIPPETS_VERSION': None,
        'MITO_CONFIG_CODE_SNIPPETS_URL': None,
        'MITO_CONFIG_DISABLE_TOURS': None
    }
    """
    return {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: mec[MITO_CONFIG_SUPPORT_EMAIL],
        MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL: None,
        MITO_CONFIG_CODE_SNIPPETS_VERSION: None,
        MITO_CONFIG_CODE_SNIPPETS_URL: None,
        MITO_CONFIG_DISABLE_TOURS: None
    }

"""
When updating the MEC_VERSION, add a function here to update the previous mec to the new version. For example, 
if mec_version='3', mec_upgrade_functions should look like:
{
   '1': upgrade_mec_1_to_2,
   '2': upgrade_mec_2_to_3
}
To keep things simple for now, these upgrade functions just make sure that all of the keys are defined so 
that the functions below can set the correct default values and format the mec properly.
"""

mec_upgrade_functions: Dict[str, Any] = {
    '1': upgrade_mec_1_to_2
}

def upgrade_mito_enterprise_configuration(mec: Optional[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
    if mec is None:
        return None

    # So mypy tests recognize that mec is not None
    _mec = mec
    while _mec[MITO_CONFIG_VERSION] in mec_upgrade_functions:
        _mec = mec_upgrade_functions[_mec[MITO_CONFIG_VERSION]](_mec)

    return _mec

def is_env_variable_set_to_true(env_variable: Union[str, int, bool]) -> bool: 
    truth_values = ['True', 'true', True, 1, '1', 't', 'T']
    return env_variable in truth_values

# Since Mito needs to look up individual environment variables, we need to 
# know the names of the variables associated with each mito config version. 
# To do so we store them as a list here. 
MEC_VERSION_KEYS = {
    '1': [MITO_CONFIG_VERSION, MITO_CONFIG_SUPPORT_EMAIL],
    '2': [
        MITO_CONFIG_VERSION, 
        MITO_CONFIG_SUPPORT_EMAIL, 
        MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL, 
        MITO_CONFIG_CODE_SNIPPETS_VERSION,
        MITO_CONFIG_CODE_SNIPPETS_URL, 
        MITO_CONFIG_DISABLE_TOURS
    ]
}

def create_mec_from_environment_variables() -> Optional[Dict[str, Any]]:
    """
    Creates a Mito Enterprise Config object from the environment variables
    """
    config_version = os.environ.get(MITO_CONFIG_VERSION)

    if config_version is None:
        return None

    mec: Dict[str, Any] = {}
    for key in MEC_VERSION_KEYS[config_version]:
        mec[key] = os.environ.get(key)

    return mec

class MitoConfig:
    """
    The MitoConfig class is repsonsible for reading the settings from the 
    environment variables and returning them as the most updated version of the 
    mito_enterprise_configuration object that is used by the rest of the app. 

    If the environment variables does not exist or does not set every configuration option, 
    the MitoConfig class sets defaults. 
    """
    def __init__(self):
        mec_potentially_outdated = create_mec_from_environment_variables()
        self.mec = upgrade_mito_enterprise_configuration(mec_potentially_outdated)

        if self.mec is not None:
            log('loaded_mito_enterprise_config')

    def get_version(self) -> str:
        if self.mec is None or self.mec[MITO_CONFIG_VERSION] is None:
            return '2' # NOTE: update this to be the most recent version, when we bump the version
        return self.mec[MITO_CONFIG_VERSION]

    def get_support_email(self) -> str:
        if self.mec is None or self.mec[MITO_CONFIG_SUPPORT_EMAIL] is None:
            return DEFAULT_MITO_CONFIG_SUPPORT_EMAIL
        return self.mec[MITO_CONFIG_SUPPORT_EMAIL]

    def get_disable_tours(self) -> bool:
        if self.mec is None or self.mec[MITO_CONFIG_DISABLE_TOURS] is None:
            return False

        disable_tours = is_env_variable_set_to_true(self.mec[MITO_CONFIG_DISABLE_TOURS])
        return True if disable_tours else False

    def _get_code_snippets_version(self) -> Optional[str]:
        if self.mec is None or self.mec[MITO_CONFIG_CODE_SNIPPETS_VERSION] is None:
            return None
        return self.mec[MITO_CONFIG_CODE_SNIPPETS_VERSION]

    def _get_code_snippets_url(self) -> Optional[str]:
        if self.mec is None or self.mec[MITO_CONFIG_CODE_SNIPPETS_URL] is None:
            return None
        return self.mec[MITO_CONFIG_CODE_SNIPPETS_URL]

    def _get_code_snippets_support_email(self) -> Optional[str]:
        if self.mec is None or self.mec[MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL] is None:
            return None
        return self.mec[MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL]

    def get_code_snippets(self) -> Optional[CodeSnippetEnvVars]:
        code_snippets_version = self._get_code_snippets_version()
        code_snippets_url = self._get_code_snippets_url()
        code_snippets_support_email = self._get_code_snippets_support_email()

        if code_snippets_version is None and code_snippets_url is None:
            return None

        if code_snippets_version != '1':
            log('mito_config_error', {'mito_config_error_reason': 'mito_config_code_snippet_version not set'})
            raise ValueError(
                "The code snippet environment variables are configured improperly. The MITO_CONFIG_CODE_SNIPPETS_URL environment variable is set, but the MITO_CONFIG_CODE_SNIPPETS_VERSION environment variable is not set to '1'.)"
            )

        if code_snippets_url is None:
            log('mito_config_error', {'mito_config_error_reason': 'mito_config_code_snippets_url not set'})
            raise ValueError(
                "The code snippet environment variables are configured improperly. The MITO_CONFIG_CODE_SNIPPETS_VERSION environment variable is set, but the MITO_CONFIG_CODE_SNIPPETS_URL environment variable is not set."
            )  

        code_snippets: CodeSnippetEnvVars = {
            'MITO_CONFIG_CODE_SNIPPETS_VERSION': code_snippets_version,
            'MITO_CONFIG_CODE_SNIPPETS_URL': code_snippets_url, 
            'MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL': code_snippets_support_email 
        }
        return code_snippets

    # Add new mito configuration options here ...

    def get_mito_config(self) -> Dict[str, Any]:
        return {
            MITO_CONFIG_VERSION: self.get_version(),
            MITO_CONFIG_SUPPORT_EMAIL: self.get_support_email(),
            MITO_CONFIG_DISABLE_TOURS: self.get_disable_tours(),
            MITO_CONFIG_CODE_SNIPPETS: self.get_code_snippets()
        }




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.


from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

if TYPE_CHECKING:
    from mitosheet.state import State
else:
    State = Any

class CodeChunk:
    """
    A CodeChunk is the a abstract base class that can be inhereited from
    to create a code generating object. For example, an AddColumnCodeChunk
    with the correct parameters and states will generate code that adds
    a column with a specific column header to a specific dataframe.

    When a step is transpiled, it generates a list of code chunks. In turn,
    we can optimize these code chunks to allow us to write the most minimal
    amount of code. 
    """

    def __init__(self, 
        prev_state: State,
        post_state: State,
    ):
        self.prev_state = prev_state
        self.post_state = post_state

    def __repr__(self) -> str:
        return self.__class__.__name__

    def get_display_name(self) -> str:
        """Returns a short name to display for this CodeChunk"""
        raise NotImplementedError('Implement in subclass')
    
    def get_description_comment(self) -> str:
        """Returns a detailed comment explaing what happened in this CodeChunk"""
        raise NotImplementedError('Implement in subclass')

    def get_code(self) -> Tuple[List[str], List[str]]:
        """
        Returns a Tuple of lists of code strings that this code chunk executes. 
        The first list is the code used to perfrom the operation. The second list 
        is the imports required for the code. ie: import pandas as pd
        """
        raise NotImplementedError('Implement in subclass')

    # TODO: we could add a function that returns a list of params and execution
    # data that you're allowed to reference, and then check this when you create
    # the step, for strong typing!

    def params_match(self, other_code_chunk: "CodeChunk", param_keys: List[str]) -> bool:
        """
        Given a different code chunk, and a list of keys to check, returns True if
        all the given keys match in the params. A useful utility for checking if 
        CodeChunks are compatible for combination.
        """
        for key in param_keys:
            if self.__dict__[key] != other_code_chunk.__dict__[key]:
                return False
        return True

    def get_created_sheet_indexes(self) -> Optional[List[int]]:
        """
        Dataframe deletes allow us to optimize a lot of code, we allow steps to
        optionally say that they only create some specific list of sheet_indexes.

        If this function returns a sheet index, and later dataframe delete steps
        delete this sheet index, then we will optimize out this step as well as
        the deleting of the dataframe.

        NOTE: if this funciton returns None, it is this CodeChunk saying that
        it cannot do any optimization with dataframe delete - which we do by
        default.
        """
        return None
    
    def get_edited_sheet_indexes(self) -> Optional[List[int]]:
        """
        Dataframe deletes allow us to optimize a lot of code, we allow steps to
        optionally say that they only edit some specific list of sheet_indexes.

        This allows us to easily optimize out these steps if the dataframe they 
        are editing is then deleted.

        NOTE: if this funciton returns None, it is this CodeChunk saying that
        it cannot do any optimization with dataframe delete - which we do by
        default.
        """
        return None
    
    def combine_right(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        """
        Given a list of CodeChunks [A, B], combine right called on A with
        B as a parameter will check if A and B can be combined into a new
        CodeChunk. 

        If they cannot be combined, None will be returned. If they can be
        combined, the new combined CodeChunk will be returned, and thus
        [A, B] goes to [A.combine_right(B)]
        """
        return None
    
    def combine_left(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        """Given a list of CodeChunks [A, B], combine right called on B with
        A as a parameter will check if A and B can be combined into a new
        CodeChunk. 

        If they cannot be combined, None will be returned. If they can be
        combined, the new combined CodeChunk will be returned, and thus
        [A, B] goes to [B.combine_right(A)]

        NOTE: combine_lefts are only done after the combine_rights. You might want
        to do one vs. the other depending on how natural it is to express. For example,
        because deleting a dataframe delete can remove a ton of other steps, expressing
        it as a combine_left is much more natural as it results in this optimization
        code all living in one location.
        """
        return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.column_steps.delete_column_code_chunk import \
    DeleteColumnsCodeChunk
from mitosheet.errors import raise_error_if_column_ids_do_not_exist
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import ColumnID


class DeleteColumnStepPerformer(StepPerformer):
    """"
    A delete_column step, which allows you to delete a column
    from a dataframe.
    """
    @classmethod
    def step_version(cls) -> int:
        return 3

    @classmethod
    def step_type(cls) -> str:
        return 'delete_column'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_ids: List[ColumnID] = get_param(params, 'column_ids')

        raise_error_if_column_ids_do_not_exist(
            'delete column',
            prev_state,
            sheet_index,
            column_ids
        )

        # Make a post state, that is a deep copy
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])

        # Actually delete the columns and update state
        post_state, pandas_processing_time = delete_column_ids(post_state, sheet_index, column_ids)

        return post_state, {
            # Add the num_cols_deleted to the execution data for logging purposes. 
            'num_cols_deleted': len(column_ids),
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            DeleteColumnsCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'), 
                get_param(params, 'column_ids'), 
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}

def delete_column_ids(
    state: State,
    sheet_index: int,
    column_ids: List[ColumnID],
) -> Tuple[State, float]:

    # Delete each column one by one
    pandas_processing_time = 0.0
    for column_id in column_ids:
        state, partial_pandas_processing_time = _delete_column_id(state, sheet_index, column_id)
        pandas_processing_time += partial_pandas_processing_time

    return state, pandas_processing_time


def _delete_column_id( 
    state: State,
    sheet_index: int,
    column_id: ColumnID
) -> Tuple[State, float]:
    
    column_header = state.column_ids.get_column_header_by_id(sheet_index, column_id)
        
    # Actually drop the column
    df = state.dfs[sheet_index]
    partial_pandas_start_time = perf_counter()
    df.drop(column_header, axis=1, inplace=True)
    partial_pandas_processing_time = perf_counter() - partial_pandas_start_time

    # And then update all the state variables removing this column from the state
    del state.column_formulas[sheet_index][column_id]
    if column_id in state.df_formats[sheet_index]['columns']:
        del state.df_formats[sheet_index]['columns'][column_id]

    # Clean up the IDs
    state.column_ids.delete_column_id(sheet_index, column_id)
    
    return state, partial_pandas_processing_time



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Changes a specific field in the user.json file
"""

from typing import Dict, List, Optional

from mitosheet.types import StepsManagerType
from mitosheet.user import set_user_field
from mitosheet.user.db import get_user_field
from mitosheet.user.schemas import UJ_RECEIVED_CHECKLISTS

CHECKLIST_UPDATE_EVENT = 'checklist_update'
CHECKLIST_UPDATE_PARAMS = ['checklist_id', 'completed_items', 'clear_other_items']


def execute_checklist_update(steps_manager: StepsManagerType, checklist_id: str, completed_items: List[str], clear_other_items: bool) -> None:
    """
    The function responsible for setting the checklist items
    in the user.json

    If clear_other_items is True, then all other items in the
    checklist will be cleared. Otherwise, the completed items
    will be appended to this list
    """
    received_checklists: Optional[Dict[str, List[str]]] = get_user_field(UJ_RECEIVED_CHECKLISTS)

    if received_checklists is None:
        received_checklists = {}

    current_completed_items = received_checklists.get(checklist_id, [])
    if clear_other_items:
        current_completed_items = completed_items
    else:
        for completed_item in completed_items:
            if completed_item not in current_completed_items:
                current_completed_items.append(completed_item)

    received_checklists[checklist_id] = current_completed_items

    set_user_field(UJ_RECEIVED_CHECKLISTS, received_checklists)

CHECKLIST_UPDATE = {
    'event_type': CHECKLIST_UPDATE_EVENT,
    'params': CHECKLIST_UPDATE_PARAMS,
    'execute': execute_checklist_update
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import base64
import io
from typing import Any, Dict

import pandas as pd
from mitosheet.excel_utils import get_df_name_as_valid_sheet_name
from mitosheet.pro.download.formatting import add_formatting_to_excel_sheet
from mitosheet.types import StepsManagerType
from mitosheet.user import is_pro


def get_dataframe_as_excel(params: Dict[str, Any], steps_manager: StepsManagerType) -> str:
    """
    Sends a dataframe as a excel string.
    """
    sheet_indexes = params['sheet_indexes']
    is_pro_user = is_pro()

    # We write to a buffer so that we don't have to save the file
    # to the file system for no reason
    buffer = io.BytesIO()
    with pd.ExcelWriter(buffer) as writer:
        for sheet_index in sheet_indexes:
            # First, write the tab in the Excel file
            df = steps_manager.dfs[sheet_index]
            df_name = steps_manager.curr_step.df_names[sheet_index]
            df.to_excel(
                writer, 
                sheet_name=get_df_name_as_valid_sheet_name(df_name),
                index=False
            )

            # Adding formatting is a pro feature
            if is_pro_user:
                add_formatting_to_excel_sheet(writer, steps_manager, sheet_index)
    
    # Go back to the start of the buffer
    buffer.seek(0)
    
    # First, we take the buffer, and base64 encode it in bytes,
    # and then we covert this to ASCII. On the front-end, we 
    # turn it back into base64, then back to bytes, before 
    # creating a Blob out of it
    return base64.b64encode(buffer.read()).decode('ascii')



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Saves the analysis with the passed name.
"""

from mitosheet.saved_analyses import write_analysis

SAVE_ANALYSIS_UPDATE_EVENT = 'save_analysis_update'
SAVE_ANALYSIS_UPDATE_PARAMS = ['analysis_name']

def execute_save_analysis_update(
        steps_manager,
        analysis_name
    ):
    """
    Saves the analysis with the passed name
    """
    write_analysis(steps_manager, analysis_name)


SAVE_ANALYSIS_UPDATE = {
    'event_type': SAVE_ANALYSIS_UPDATE_EVENT,
    'params': SAVE_ANALYSIS_UPDATE_PARAMS,
    'execute': execute_save_analysis_update
} 


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import deepcopy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import pandas as pd

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.column_steps.set_column_formula_code_chunk import \
    SetColumnFormulaCodeChunk
from mitosheet.errors import (MitoError, make_execution_error,
                              make_operator_type_error,
                              make_unsupported_function_error,
                              raise_error_if_column_ids_do_not_exist)
from mitosheet.parser import get_frontend_formula, parse_formula
from mitosheet.sheet_functions import FUNCTIONS
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import FORMULA_ENTIRE_COLUMN_TYPE, ColumnHeader, ColumnID, FormulaAppliedToType


class SetColumnFormulaStepPerformer(StepPerformer):
    """
    A set_column_formula step, which allows you to set the formula
    of a given column in the sheet (and then recalculates this column)
    and it's dependents.
    """

    @classmethod
    def step_version(cls) -> int:
        return 4

    @classmethod
    def step_type(cls) -> str:
        return 'set_column_formula'

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any]) -> Dict[str, Any]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_id: ColumnID = get_param(params, 'column_id')
        formula_label: Union[str, bool, int, float] = get_param(params, 'formula_label')
        index_labels_formula_is_applied_to: FormulaAppliedToType = get_param(params, 'index_labels_formula_is_applied_to')
        new_formula: str = get_param(params, 'new_formula')

        column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)
        
        # If the user submits an empty formula, we just set it equal to zero
        if new_formula == '':
            params['new_formula'] = '=0'
        else:
            try:
                # Try and parse the formula, letting it throw errors if it
                # is invalid
                parse_formula(new_formula, column_header, formula_label, index_labels_formula_is_applied_to, prev_state.dfs[sheet_index], throw_errors=True)
            except Exception as e:                
                params['new_formula'] = _get_fixed_invalid_formula(new_formula, column_header, formula_label, index_labels_formula_is_applied_to, prev_state.dfs[sheet_index])

        return params

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_id: ColumnID = get_param(params, 'column_id')
        formula_label: Union[str, bool, int, float] = get_param(params, 'formula_label')
        index_labels_formula_is_applied_to: FormulaAppliedToType = get_param(params, 'index_labels_formula_is_applied_to')
        new_formula: str = get_param(params, 'new_formula')

        raise_error_if_column_ids_do_not_exist(
            'set column formula',
            prev_state,
            sheet_index,
            column_id
        )

        column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)

        # Then we try and parse the formula
        _, new_functions, _, _ = parse_formula(
            new_formula, 
            column_header,
            formula_label,
            index_labels_formula_is_applied_to,
            prev_state.dfs[sheet_index]
        )

        # The formula can only reference known formulas
        missing_functions = new_functions.difference(set(FUNCTIONS.keys()))
        if any(missing_functions):
            raise make_unsupported_function_error(missing_functions, error_modal=False)

        # We check out a new step
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])

        # Update the column formula, and then execute the new formula graph
        try:
            pandas_start_time = perf_counter()
            exec_column_formula(post_state, post_state.dfs[sheet_index], sheet_index, column_id, formula_label, index_labels_formula_is_applied_to, new_formula)
            pandas_processing_time = perf_counter() - pandas_start_time
        except MitoError as e:
            # Catch the error and make sure that we don't set the error modal
            e.error_modal = False
            raise e
        except Exception as e:
            raise make_execution_error(error_modal=False)

        return post_state, {
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        """
        Transpiles an set_column_formula step to python code!
        """
        return [
            SetColumnFormulaCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_id'),
                get_param(params, 'formula_label'),
                get_param(params, 'index_labels_formula_is_applied_to'),
                get_param(params, 'new_formula'),
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}


def _get_fixed_invalid_formula(
        new_formula: str, 
        column_header: ColumnHeader, 
        formula_label: Union[str, bool, int, float],
        index_labels_formula_is_applied_to: FormulaAppliedToType,
        df: pd.DataFrame
    ) -> str:
    """
    A helper function that, given a formula, will try and fix
    any common errors with it. 

    Currently, the common errors are just checking if the formula
    is missing some number of parenthases at the end.

    Returns the fixed formula. If it cannot fix the formula, then 
    returns the original formula (so that execution continues as 
    normal, to report the error to the user).
    """
    POTENTIAL_VALID_FORMULAS = [
        # Check the original formula, just in case it actually is valid
        new_formula,
        new_formula + ')',
        new_formula + '))',
        new_formula + ')))',
    ]
    for fixed_formula in POTENTIAL_VALID_FORMULAS:
        try:
            # Parse the formula, and return if it is valid
            parse_formula(fixed_formula, column_header, formula_label, index_labels_formula_is_applied_to, df, throw_errors=True)
            return fixed_formula
        except:
            pass
    
    return new_formula


def get_details_from_operator_type_error(error: TypeError) -> Optional[Tuple[str, str, str]]:
    """
    We detect operator errors by checking the error string, which has the format:

    If you write 1 + 'true'
    - unsupported operand type(s) for /: 'int' and 'str' 

    If you write 1 + pd.Series(['true'])
    - can only concatenate str (not "int") to str

    If you write 10 ^ pd.Series(['true'])
    - Cannot perform 'rxor' with a dtyped [object] array and scalar of type [bool]
    If you write pd.Series(['true']) ^ 10
    - Cannot perform 'xor' with a dtyped [object] array and scalar of type [bool]

    On Python 3.6, there is one error when you try and append a number to 
    a string, where it says `must be str, not int`.

    ^ NOTE: the above two errors makes pretty much no sense at all, but it appears
    to be casting the number to a boolean? So we choose to handle this as a number,
    as this is when you would be 

    Returns None if the passed error is not an operator type error. Otherwise, returns
    a triple of (operator, first argument type, second argument type).
    """
    error_message = str(error)
    # Handle case where standard python primitive types are used incorrectly
    if error_message.startswith('unsupported operand type'):
        # Then, we get the specific operator by checking right before the :
        operator = error_message.split(':')[0].split(' ')[-1].strip()
        # And we get the specific types - see above format for how this parsing works!
        arg_one_type = error_message.split(':')[-1].split('\'')[1]
        arg_two_type = error_message.split(':')[-1].split('\'')[-2]
        return (operator, arg_one_type, arg_two_type)
    # Handle case where pandas dataframes are part of the operator, for a concatenate
    if error_message.startswith('can only concatenate str'):
        # We cannot tell the order here (e.g. we don't know if a string was concated
        # to an integer, or the other way around) b/c the error doesn't say so, so we just
        # do our best 
        other_type = error_message.split('"')[1]
        return ('+', 'str', other_type)
    # Handle a ^
    if error_message.startswith('Cannot perform \'rxor\' with a dtyped') or error_message.startswith('Cannot perform \'xor\' with a dtyped'):
        # This error message might be totally wrong, but there is some weird
        # casting that goes on in the error message that makes it hard to tell. As such,
        # we report the most likely error.
        return ('^', 'number', 'str')
    if error_message.startswith('must be str, not'):
        # When adding a number to a string
        return ('+', 'str', 'number')

    return None


def exec_column_formula(
    post_state: State, 
    df: pd.DataFrame, 
    sheet_index: int, 
    column_id: ColumnID, 
    formula_label: Union[str, bool, int, float], 
    index_labels_formula_is_applied_to: FormulaAppliedToType,
    spreadsheet_code: str
) -> None:
    """
    Helper function for refreshing the column when the formula is set
    """

    df_name = post_state.df_names[sheet_index]

    if spreadsheet_code == '':
        return

    column_header = post_state.column_ids.get_column_header_by_id(sheet_index, column_id)
    python_code, _, _, _ = parse_formula(
        spreadsheet_code, 
        column_header,
        formula_label,
        index_labels_formula_is_applied_to,
        post_state.dfs[sheet_index]
    )

    try:
        # Exec the code, where the df is the original dataframe
        # See explination here: https://www.tutorialspoint.com/exec-in-python
        exec(
            python_code,
            {'df': df, 'pd': pd}, 
            FUNCTIONS
        )
        # Then, update the column spreadsheet code
        frontend_formula = get_frontend_formula(
            spreadsheet_code,
            formula_label,
            post_state.dfs[sheet_index]
        )
        
        # If the user is setting the entire column, then there is only one formula for every cell in
        # the entire column. But if they are just setting specific indexes, we need to store the formulas
        # before this as well, so that we can figure out what formula is applied to each index
        if index_labels_formula_is_applied_to['type'] == FORMULA_ENTIRE_COLUMN_TYPE:
            post_state.column_formulas[sheet_index][column_id] = [{'frontend_formula': frontend_formula, 'location': index_labels_formula_is_applied_to, 'index': df.index.to_list()}]
        else:
            post_state.column_formulas[sheet_index][column_id].append({'frontend_formula': frontend_formula, 'location': index_labels_formula_is_applied_to, 'index': df.index.to_list()})

    except TypeError as e:
        # We catch TypeErrors specificially, so that we can case on operator errors, to 
        # give better error messages
        operator_type_error_details = get_details_from_operator_type_error(e)
        if operator_type_error_details is not None:
            # If there is an operator error, we handle it specially, to give the user
            # more information about how to recover
            raise make_operator_type_error(*operator_type_error_details)
        else:
            # If it's not an operator error, we just propagate the error up
            raise e
    except NameError as e:
        # If we have a column header that does not exist in the formula, we may
        # throw a name error, in which case we alert the user
        column_header = str(e).split('\'')[1]
        raise MitoError(
            'no_column_error',
            'No Column Exists',
            f'Setting a column formula failed. The column "{str(column_header)}" referenced in the formula does not exist in {df_name}.',
            error_modal=True
        )
    except Exception as e:
        raise



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
This folder contains all of the functions that upgrade
specific steps when the format changes.

Each upgrade function takes two inputs: the step to update,
and all steps that come after it. Each function returns a 
list of steps, which represent the now upgraded step and the
steps that come after it (that may have been changed in some
way, per the upgrade).

Thus, the upgrade function can transform the step whose format
changed, but also modify the steps that come later, if they
need to changed to make them consistent with the upgrade.
"""


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from abc import ABC, abstractmethod
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from typing import Any, Dict, List, Optional, Set, Tuple
 
class StepPerformer(ABC, object):
    """
    The abstract base class for a step performer, which are the set
    of classes that execute, transpile, and describe steps. 
    """

    @classmethod
    @abstractmethod
    def step_version(cls) -> int:
        """
        Returns the version of the step. Changes when the parameters
        of the step change.
        """
        pass

    @classmethod
    @abstractmethod
    def step_type(cls) -> str:
        """
        The name of the step used internally. If you change this, you must upgrade
        the step and bump the version.
        """
        pass
    
    @classmethod
    def step_event_type(cls) -> str:
        """
        The type of the edit event that generates this type of step. 
        This is always just f'{cls.step_type}_edit'.
        """
        return f'{cls.step_type()}_edit'

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any]) -> Dict[str, Any]:
        """
        Given the parameters of the step, will saturate the event with
        more parameters based on the passed prev_state. 
        """
        # By default, we don't do anything with the saturate
        return params

    @classmethod
    @abstractmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        """
        Execute always returns the post_state, and optionally returns a dictionary
        of execution_data, which is data that may be useful to the transpiler in
        transpiling the code.

        If the execution_data also includes the key `pandas_processing_time`, this 
        will allow the logging infrastructure to determine how much overhead Mito
        adds to executing this event.

        If the exectution_data includes the key `result`, than this result can be
        accessed in the frontend to display some results to the user. See the 
        DropDuplicatesStepPerformer and the DropDuplicatesCodeChunk for an example of this usage.
        """
        pass

    @classmethod
    @abstractmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        """
        Returns a list of the CodeChunks that correspond to this 
        step being executed
        """
        pass

    @classmethod
    @abstractmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        """
        Returns a set of all the sheet indexes that were modified
        by this step.

        If it returns an empty set, then this step
        modified every dataframe.

        If it returned -1, then it modified all new dataframes (on
        the left side of the dfs array).
        """
        pass


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Set, Tuple, Union

import pandas as pd
import numpy as np
from mitosheet.types import ColumnHeader
from mitosheet.utils import is_prev_version

# TAB is used in place of \t in generated code because
# Jupyter turns \t into a grey arrow, but converts four spaces into a tab.
TAB = '    '
NEWLINE_TAB = f'\n{TAB}'
NEWLINE = '\n'


def column_header_list_to_transpiled_code(column_headers: Union[List[ColumnHeader], Set[ColumnHeader], List[Tuple[str, Optional[str]]]]) -> str:
    """
    A helper function for turning a list of column headers into a 
    valid list of Python code.
    """
    transpiled_column_headers = [
        column_header_to_transpiled_code(column_header)
        for column_header in column_headers
    ]
    joined_transpiled_column_headers = ', '.join(transpiled_column_headers)
    return f'[{joined_transpiled_column_headers}]'


def column_header_to_transpiled_code(column_header: ColumnHeader) -> str:
    """
    Makes sure the column header is correctly transpiled to 
    code in a way that makes sure it's referenced properly.

    Handles multi-index, boolean, string, and number columns 
    correctly.
    """
    # If this is a multi-index header, then we turn each of the pieces of the column
    # header into valid transpiled code, and then we combine them into a tuple
    if isinstance(column_header, tuple):
        column_header_parts = [column_header_to_transpiled_code(column_header_part) for column_header_part in column_header]
        column_header_parts_joined = ', '.join(column_header_parts)
        return f'({column_header_parts_joined})'

    # We must handle np.nan first because isinstance(np.nan, float) evaluates to True
    if not is_prev_version(pd.__version__, '1.0.0') and column_header is np.nan:
        return 'pd.np.nan'
    elif isinstance(column_header, int) or isinstance(column_header, float) or isinstance(column_header, bool):
        return str(column_header)
    elif isinstance(column_header, pd.Timestamp):
        return f'pd.to_datetime(\'{column_header.strftime("%Y-%m-%d %X")}\')'
    elif isinstance(column_header, pd.Timedelta):
        return f'pd.to_timedelta(\'{str(column_header)}\')'
    elif column_header is pd.NaT:
        return 'pd.NaT'
    elif not is_prev_version(pd.__version__, '1.0.0') and column_header is pd.NA:
        return 'pd.NA'
        
    return repr(column_header)

def list_to_string_without_internal_quotes(list: List[Any]) -> str:
    """
    Helper function for formatting a list as a string without 
    leading and trailing '
    """
    string = (', ').join(list)
    return "[" + string +  "]"

def column_header_map_to_string(column_header_map: Dict[ColumnHeader, ColumnHeader]) -> str:
    if len(column_header_map) <= 3:
        # If there are only a few column headers, we put them in a single line
        result = '{' 
        for column_header_key, column_header_value in column_header_map.items():
            result += f'{column_header_to_transpiled_code(column_header_key)}: {column_header_to_transpiled_code(column_header_value)}, '
        result = result[:-2] + "}" # don't take the last comma and space
        return result
    else:
        result = '{\n' 
        for column_header_key, column_header_value in column_header_map.items():
            result += f'{TAB}{column_header_to_transpiled_code(column_header_key)}: {column_header_to_transpiled_code(column_header_value)},\n'
        result = result[:-2] + "\n}" # don't take the last comma and new line
        return result


def param_dict_to_code(param_dict: Dict[str, Any], level: int=0, as_single_line: bool=False) -> str:
    """
    Takes a potentially nested params dictonary and turns it into a
    code string that we can use in the graph generated code.

    level should be 0 if we are at the highest level dict, and otherwise
    should increment by 1 anytime we enter a new subdictonary.
    """

    # Make sure we handle as a single line properly
    if as_single_line:
        TAB_CONSTANT = ''
        NEWLINE_CONSTANT = ''
    else:
        TAB_CONSTANT = TAB
        NEWLINE_CONSTANT = '\n'

    if level == 0:
        code = f"{NEWLINE_CONSTANT}"
    else:
        code = f"dict({NEWLINE_CONSTANT}"

    value_num = 0
    for key, value in param_dict.items():
        if isinstance(value, dict):
            # Recurse on this nested param dictonary
            code_chunk = f"{key} = {param_dict_to_code(value, level=level + 1)}"
        else:
            # We use this slighly misnamed function to make sure values get transpiled right
            code_chunk = f"{key}={column_header_to_transpiled_code(value)}"
        
        # If we're not on the first value in this dict, we need to add a 
        # command new line after the last value
        if value_num != 0:
            code += f", {NEWLINE_CONSTANT}"

        value_num += 1

        # Add spacing before the param
        code += f"{TAB_CONSTANT * (level + 1)}"

        code += f"{code_chunk}"

    if level == 0:
        code += f"{NEWLINE_CONSTANT}"
    else:
        # Make sure to close the dict
        code += f"{NEWLINE_CONSTANT}{TAB_CONSTANT * (level)})"
    
    return code


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

"""
The mitosheet package is distributed under the mitosheet and mitosheet3
package names on pip. The package.json will tell you all you need to 
know about which one we are in currently using.

As such, this setup.py script reads in the package.json and sets up
the proper package.
"""

from __future__ import print_function
from glob import glob
from os.path import join as pjoin
import json
import setuptools
from pathlib import Path


from jupyter_packaging import (
    create_cmdclass,
    install_npm,
    ensure_targets,
    combine_commands,
    skip_if_exists,
)

from setuptools import setup

HERE = Path(__file__).parent.resolve()

package_json = json.loads(open('package.json').read())
lab_path = Path(pjoin(HERE, 'mitosheet', 'labextension'))
notebook_path = Path(pjoin(HERE, 'mitosheet', 'nbextension'))

python_requires='>=3.4'

# The name of the project
name = package_json['name']

if name == 'mitosheet2':
    # Get our version, which we just read 
    version = package_json['version']

    # Representative files that should exist after a successful build
    jstargets = [
        pjoin(HERE, 'lib', 'plugin.js'),
    ]

    package_data_spec = {
        name: [
            'labextension/*.tgz'
        ]
    }

    data_files_spec = [
        ('share/jupyter/lab/extensions', lab_path, '*.tgz'),
    ]


    cmdclass = create_cmdclass('jsdeps', package_data_spec=package_data_spec,
        data_files_spec=data_files_spec)
    cmdclass['jsdeps'] = combine_commands(
        install_npm(HERE, build_cmd='build:all'),
        ensure_targets(jstargets),
    )

    setup_args = dict(
        name            = name,
        description     = 'The Mito Spreadsheet',
        version         = version,
        scripts         = glob(pjoin('scripts', '*')),
        cmdclass        = cmdclass,
        packages        = setuptools.find_packages(exclude=['deployment']),
        author          = 'Mito',
        author_email    = 'naterush1997@gmail.com',
        license         = "GNU Affero General Public License v3",
        url             = 'https://github.com/mito-ds/monorepo',
        platforms       = "Linux, Mac OS X, Windows",
        keywords        = ['Jupyter', 'Widgets', 'IPython'],
        classifiers     = [
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            "License :: OSI Approved :: GNU Affero General Public License v3",
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Framework :: Jupyter',
        ],
        include_package_data = True,
        # NOTE: this should be the same as the INSTALL_REQUIRES variable in
        # the mitoinstaller package
        install_requires = [
            # We require jupyterlab 2.0
            'jupyterlab>=2.0,<3.0,!=2.3.0,!=2.3.1', # there are css incompatabilities on version 2.3.1 and 2.3.0
            'ipywidgets~=7.0.0',
            # We allow users to have many versions of pandas installed. All functionality should
            # work, with the exception of Excel import, which might require additonal dependencies
            'pandas>=0.24.2',
            # We don't need to lock an analytics-python version, as this library
            # is stabilized and mature
            'analytics-python',
            # Graphing libraries
            'plotly>=4.14.3',
            'chardet>=3.0.4',
            # For XLSX, reading - we don't fix so works on all python versions
            'openpyxl',
            # xlsxwriter is needed for adding formatting to exported Excel sheets. 
            # We pin to a pretty old version because the formatting functionality hasn't changed in a long time.
            'xlsxwriter>=0.6.9,<=3.0.2 '
        ],
        extras_require = {
            'test': [
                'pytest',
                'flake8',
                'types-chardet',
                'types-requests',
                'mypy',
                'pytest_httpserver'
            ],
            'deploy': [
                'wheel', 
                'twine',
                "jupyter_packaging<=0.10.6",
                "setuptools==56.0.0"
            ],
            'optional_feature_dependencies': [
                # According to this documentation (https://github.com/snowflakedb/snowflake-connector-python),
                # snowflake-connect-python requires at least Python 3.7
                'snowflake-connector-python[pandas]; python_version>="3.7"',
            ]
        },
        entry_points = {
        },
        long_description="""
            To learn more about Mito, checkout out our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
            Before installing Mito \n\n
            1. Check that you have Python 3.6 or above. To check your version of Python, open a new terminal, and type python3 --version. If you need to install or update Python, restart your terminal after doing so.\n\n
            2. Check that you have Node installed.To check this, open a new terminal, and type node -v.  It should print a version number. If you need to install Node, restart your terminal after doing so.\n\n
            3. Mito works in Jupyter Lab 2.0 only. We do not yet support Google Collab, VSCode, or Jupyter Lab 3.0.\n\n
            4. Checkout our terms of service and privacy policy. By installing Mito, you're agreeing to both of them. Please contact us at aarondr77 (@) gmail.com with any questions.\n\n
            Installation Instructions \n\n
            For more detailed installation instructions, see our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
            1. pip install mitoinstaller\n\n
            2. python -m mitoinstaller install
            """,
        long_description_content_type='text/markdown'
    )
elif name == 'mitosheet' or name == 'mitosheet3' or name == 'mitosheet-private':
    # Representative files that should exist after a successful build
    jstargets = [
        str(lab_path / "package.json"),
        str(notebook_path / "index.js"),
    ]

    package_data_spec = {
        'mitosheet': ["*"],
    }

    labext_name = name

    data_files_spec = [
        # Notebook extension data files
        ('share/jupyter/nbextensions/mitosheet', notebook_path, '**'),
        ('etc/jupyter/nbconfig/notebook.d', '.', 'mitosheet.json'),

        # Lab extension data files
        ("share/jupyter/labextensions/%s" % labext_name, str(lab_path), "**"),
        ("share/jupyter/labextensions/%s" % labext_name, str(HERE), "install.json"),
    ]


    cmdclass = create_cmdclass("jsdeps",
        package_data_spec=package_data_spec,
        data_files_spec=data_files_spec
    )

    js_command = combine_commands(
        install_npm(HERE, build_cmd="build:all", npm=["jlpm"]),
        ensure_targets(jstargets),
    )

    is_repo = (HERE / ".git").exists()
    if is_repo:
        cmdclass["jsdeps"] = js_command
    else:
        cmdclass["jsdeps"] = skip_if_exists(jstargets, js_command)

    setup_args = dict(
        name                    = name,
        version                 = package_json["version"],
        url                     = package_json["homepage"],
        author                  = package_json["author"]["name"],
        author_email            = package_json["author"]["email"],
        description             = package_json["description"],
        license                 = "GNU Affero General Public License v3",
        long_description="""
            To learn more about Mito, checkout out our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
            Before installing Mito \n\n
            1. Check that you have Python 3.6 or above. To check your version of Python, open a new terminal, and type python3 --version. If you need to install or update Python, restart your terminal after doing so.\n\n
            2. Checkout our terms of service and privacy policy. By installing Mito, you're agreeing to both of them. Please contact us at aarondr77 (@) gmail.com with any questions.\n\n
            Installation Instructions \n\n
            For more detailed installation instructions, see our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
            1. pip install mitoinstaller\n\n
            2. python -m mitoinstaller install
        """,
        long_description_content_type = "text/markdown",
        cmdclass                 = cmdclass,
        packages                 = setuptools.find_packages(exclude=['deployment']),
        install_requires=[        
            "jupyterlab~=3.0",
            # We allow users to have many versions of pandas installed. All functionality should
            # work, with the exception of Excel import, which might require additonal dependencies
            'pandas>=0.24.2',
            'analytics-python',
            # Graphing libraries
            'plotly>=4.14.3',
            'chardet>=3.0.4',
            # For XLSX, reading - we don't fix so works on all python versions
            'openpyxl',
            # xlsxwriter is needed for adding formatting to exported Excel sheets. 
            # We pin to a pretty old version because the formatting functionality hasn't changed in a long time.
            'xlsxwriter>=0.6.9,<=3.0.2'
        ],
        extras_require = {
            'test': [
                'pytest',
                'flake8',
                'types-chardet',
                'types-requests',
                'mypy',
                'pytest_httpserver'
            ],
            'deploy': [
                'wheel', 
                'twine',
                "jupyter_packaging<=0.10.6",
                "setuptools==56.0.0"
            ],
            'optional_feature_dependencies': [
                # According to this documentation (https://github.com/snowflakedb/snowflake-connector-python),
                # snowflake-connect-python requires at least Python 3.7
                'snowflake-connector-python[pandas]; python_version>="3.7"',
            ]
        },
        zip_safe                = False,
        include_package_data    = True,
        python_requires         = ">=3.6",
        platforms               = "Linux, Mac OS X, Windows",
        keywords                = ["Jupyter", "JupyterLab", "JupyterLab3"],
        classifiers             = [
            "License :: OSI Approved :: GNU Affero General Public License v3",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Framework :: Jupyter",
        ],
    )

if __name__ == '__main__':
    setup(**setup_args)




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.promote_row_to_header_code_chunk import PromoteRowToHeaderCodeChunk
from mitosheet.errors import make_invalid_promote_row_to_header

from mitosheet.state import State
from mitosheet.step_performers.column_steps.rename_column import rename_column_headers_in_state
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.utils import MAX_ROWS, convert_df_to_parsed_json

class PromoteRowToHeaderStepPerformer(StepPerformer):
    """
    Allows you to promote a row to a header (and deletes that row in the process).
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'promote_row_to_header'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        index: Any = get_param(params, 'index')

        # We make a new state to modify it
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])

        pandas_processing_time = 0.0

        for old_column_header, new_column_header in zip(post_state.dfs[sheet_index].columns, post_state.dfs[sheet_index].loc[index]):

            column_id = post_state.column_ids.get_column_id_by_header(sheet_index, old_column_header)
            _, partial_pandas_processing_time = rename_column_headers_in_state(
                post_state,
                sheet_index,
                column_id,
                new_column_header
            )
            
            pandas_processing_time += partial_pandas_processing_time

        pandas_start_time_drop = perf_counter()
        post_state.dfs[sheet_index].drop(labels=[index], inplace=True)
        pandas_processing_time_drop = perf_counter() - pandas_start_time_drop

        # We make sure that this making of headers will work, and not cause issues 
        # later while trying to convert to json. We throw an error if this causes
        # errors. See this bug: https://github.com/mito-ds/monorepo/issues/267
        try:
            convert_df_to_parsed_json(post_state.dfs[sheet_index].head(1))
        except:
            raise make_invalid_promote_row_to_header()

        return post_state, {
            'pandas_processing_time': pandas_processing_time + pandas_processing_time_drop,
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            PromoteRowToHeaderCodeChunk(prev_state, post_state, get_param(params, 'sheet_index'), get_param(params, 'index'))
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import List

from mitosheet.step_performers.import_steps.simple_import import SimpleImportStepPerformer
from mitosheet.types import StepsManagerType

CLEAR_EVENT = 'clear'
CLEAR_PARAMS: List[str] = []

def execute_clear_update(steps_manager: StepsManagerType) -> None:
    steps_manager.execute_clear()

CLEAR_UPDATE = {
    'event_type': CLEAR_EVENT,
    'params': CLEAR_PARAMS,
    'execute': execute_clear_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from copy import copy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.fill_na_code_chunk import FillNaCodeChunk
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import ColumnID


class FillNaStepPerformer(StepPerformer):
    """
    Allows you to fill nan values.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'fill_na'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:

        sheet_index = get_param(params, 'sheet_index')
        column_ids = get_param(params, 'column_ids')
        fill_method = get_param(params, 'fill_method')
        fill_method_type = fill_method['type']

        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])

        df = post_state.dfs[sheet_index]
        column_headers = post_state.column_ids.get_column_headers_by_ids(sheet_index, column_ids)

        pandas_start_time = perf_counter()

        if fill_method_type == 'value':
            values = {column_header: fill_method['value'] for column_header in column_headers}
            df.fillna(values, inplace=True)
        elif fill_method_type == 'ffill':
            df[column_headers] = df[column_headers].fillna(method='ffill')
        elif fill_method_type == 'bfill':
            df[column_headers] = df[column_headers].fillna(method='bfill')
        elif fill_method_type == 'mean':
            df[column_headers] = df[column_headers].fillna(df[column_headers].mean(numeric_only=False))
        elif fill_method_type == 'median':
            df[column_headers] = df[column_headers].fillna(df[column_headers].median(numeric_only=False))
        else:
            raise Exception(f"Invalid fill method {fill_method}")

        pandas_processing_time = perf_counter() - pandas_start_time

        return post_state, {
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            FillNaCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_ids'),
                get_param(params, 'fill_method'),
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
Main file containing the mito widget.
"""
import json
import os
import re
import time
from sysconfig import get_python_version
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
from ipykernel import get_connection_file
from ipykernel.comm import Comm
from IPython import get_ipython
from IPython.display import HTML, display

from mitosheet.api import API
from mitosheet.data_in_mito import DataTypeInMito
from mitosheet.enterprise.mito_config import MitoConfig
from mitosheet.errors import (MitoError, get_recent_traceback,
                              make_execution_error)
from mitosheet.saved_analyses import write_analysis
from mitosheet.steps_manager import StepsManager
from mitosheet.telemetry.telemetry_utils import (log, log_event_processed,
                                                 telemetry_turned_on)
from mitosheet.updates.replay_analysis import REPLAY_ANALYSIS_UPDATE
from mitosheet.user import is_local_deployment, should_upgrade_mitosheet
from mitosheet.user.create import try_create_user_json_file
from mitosheet.user.db import USER_JSON_PATH, get_user_field
from mitosheet.user.location import is_in_google_colab, is_in_vs_code
from mitosheet.user.schemas import (UJ_MITOSHEET_LAST_FIFTY_USAGES,
                                    UJ_RECEIVED_CHECKLISTS, UJ_RECEIVED_TOURS,
                                    UJ_USER_EMAIL)
from mitosheet.user.utils import get_pandas_version, is_pro, is_running_test
from mitosheet.utils import get_new_id
from mitosheet.api.get_validate_snowflake_credentials import get_cached_snowflake_credentials


class MitoBackend():
    """
        The MitoWidget holds all of the backend state for the Mito extension, and syncs
        the state with the frontend widget. 
    """
    mito_comm: Optional[Comm] = None
    
    def __init__(self, *args: List[Union[pd.DataFrame, str]], analysis_to_replay: Optional[str]=None):
        """
        Takes a list of dataframes and strings that are paths to CSV files
        passed through *args.
        """
        # Call the DOMWidget constructor to set up the widget properly
        super(MitoBackend, self).__init__()

        # Setup the MitoConfig class
        self.mito_config = MitoConfig() # type: ignore
            
        # Set up the state container to hold private widget state
        self.steps_manager = StepsManager(args, mito_config=self.mito_config, analysis_to_replay=analysis_to_replay)

        # And the api
        self.api = API(self.steps_manager, self)

        # We store static variables to make writing the shared
        # state variables quicker; we store them so we don't 
        # have to recompute them on each update
        last_50_usages = get_user_field(UJ_MITOSHEET_LAST_FIFTY_USAGES)
        self.num_usages = len(last_50_usages if last_50_usages is not None else [])
        self.is_local_deployment = is_local_deployment()
        self.should_upgrade_mitosheet = should_upgrade_mitosheet()
        self.received_tours = get_user_field(UJ_RECEIVED_TOURS)


    @property
    def analysis_name(self):
        return self.steps_manager.analysis_name

    @property
    def mito_send(self):
        if self.mito_comm:
            return self.mito_comm.send
        else:
            # If we don't have a comm defined, this is because we are running a test, and so 
            # we simply don't do anything with messages that are tried to send. In the future, 
            # we can save them somewhere, and then make assertions about them -- cool!
            return lambda _: _

    def get_shared_state_variables(self) -> Dict[str, Any]:
        """
        Helper function for updating all the variables that are shared
        between the backend and the frontend through trailets.
        """
        return {
            'sheet_data_json': self.steps_manager.sheet_data_json,
            'analysis_data_json': self.steps_manager.analysis_data_json,
            'user_profile_json': self.get_user_profile_json()
        }

    def get_user_profile_json(self) -> str:
        return json.dumps({
            # Dynamic, update each time
            'userEmail': get_user_field(UJ_USER_EMAIL),
            'receivedTours': get_user_field(UJ_RECEIVED_TOURS),
            'receivedChecklists': get_user_field(UJ_RECEIVED_CHECKLISTS),
            'isPro': is_pro(),
            'telemetryEnabled': telemetry_turned_on(),
            # Static over a single analysis
            'pythonVersion': get_python_version(),
            'pandasVersion': get_pandas_version(),
            'isLocalDeployment': self.is_local_deployment,
            'shouldUpgradeMitosheet': self.should_upgrade_mitosheet,
            'numUsages': self.num_usages,
            'mitoConfig': self.steps_manager.mito_config.get_mito_config(),
            'snowflakeCredentials': get_cached_snowflake_credentials()
        })


    def handle_edit_event(self, event: Dict[str, Any]) -> None:
        """
        Handles an edit_event. Per the spec, an edit_event
        updates both the sheet and the codeblock, and as such
        the sheet is re-evaluated and the code for the codeblock
        is re-transpiled.

        Useful for any event that changes the state of both the sheet
        and the codeblock!
        """

        # First, we send this new edit to the evaluator
        self.steps_manager.handle_edit_event(event)

        # Also, write the analysis to a file!
        write_analysis(self.steps_manager)

        # Tell the front-end to render the new sheet and new code with an empty
        # response. NOTE: in the future, we can actually send back some data
        # with the response (like an error), to get this response in-place!        
        self.mito_send({
            'event': 'response',
            'id': event['id'],
            'shared_variables': self.get_shared_state_variables()
        })


    def handle_update_event(self, event: Dict[str, Any]) -> None:
        """
        This event is not the user editing the sheet, but rather information
        that has been collected from the frontend (after render) that is being
        passed back.

        For example:
        - Names of the dataframes
        - Name of an existing analysis
        """

        try:
            self.steps_manager.handle_update_event(event)
        except Exception as e:
            # We handle the case of replaying the analysis specially, because we don't
            # want to display the error modal - we want to display something specific
            # in this case. Note that we include the updating of shared state variables
            # in the try catch, as this is sometimes where errors occur
            if event["type"] == REPLAY_ANALYSIS_UPDATE['event_type']:
                # Propagate the Mito error if we can. We never want a replay analysis to open
                # an error modal, so make sure to update the error if necessary
                if isinstance(e, MitoError):
                    e.error_modal = False
                    raise e
                raise make_execution_error(error_modal=False)
            raise
        # Also, write the analysis to a file!
        write_analysis(self.steps_manager)

        # Tell the front-end to render the new sheet and new code with an empty
        # response. 
        self.mito_send({
            'event': 'response',
            'id': event['id'],
            'shared_variables': self.get_shared_state_variables()
        })

    def receive_message(self, content: Dict[str, Any]) -> bool:
        """
        Handles all incoming messages from the JS widget. There are three main
        types of events:

        1. edit_event: any event that updates the state of the sheet and the
        code block at once. Leads to reevaluation, and a re-transpile.

        2. update_event: any event that isn't caused by an edit, but instead
        other types of new data coming from the frontend (e.g. the df names 
        or some existing steps).

        3. api_call: an event that is used to retrieve information from the backend without
        updating the backend state.

        4. A log_event is just an event that should get logged on the backend.
        """

        start_time: Optional[float] = time.perf_counter()
        event = content

        try:
            if event['event'] == 'edit_event':
                self.handle_edit_event(event)
            elif event['event'] == 'update_event':
                self.handle_update_event(event)
            elif event['event'] == 'api_call':
                self.api.process_new_api_call(event)
                # NOTE: since API calls are in a seperate thread, their start time and end
                # time are not valid, and so we don't even log the start time to not be confusing
                start_time = None

            if event['event'] != 'api_call':
                # NOTE: we don't need to case on log_event above because it always gets
                # passed to this function, and thus is logged. We also don't log in the
                # case of an API event as that is in a seperate thread and so takes care
                # of it's logging seperately
                log_event_processed(event, self.steps_manager, start_time=start_time)

            return True
        except MitoError as e:
            if is_running_test():
                print(get_recent_traceback())
                print(e)

            # Log processing this event failed
            log_event_processed(event, self.steps_manager, failed=True, mito_error=e, start_time=start_time)

            # If the error says to ignore the error modal, then we
            # send some data with the response so that the frontend
            # knows to ignore the error moda 
            response = {
                'event': 'edit_error',
                'id': event['id'],
                'type': e.type_,
                'header': e.header,
                'to_fix': e.to_fix,
                'traceback': e.traceback,
            }
            if not e.error_modal:
                response['data'] = {
                    'event': 'edit_error',
                    'type': e.type_,
                    'header': e.header,
                    'to_fix': e.to_fix,
                    'traceback': e.traceback,
                }

            # Report it to the user, and then return
            self.mito_send(response)
        except:
            if is_running_test():
                print(get_recent_traceback())
            # We log that processing failed, but have no edit error
            log_event_processed(event, self.steps_manager, failed=True, start_time=start_time)
            # Report it to the user, and then return
            self.mito_send({
                'event': 'edit_error',
                'id': event['id'],
                'type': 'execution_error',
                'header': 'Execution Error',
                'to_fix': 'Sorry, there was an error during executing this code.',
                'traceback': get_recent_traceback()
            })

        return False

with open(os.path.normpath(os.path.join(__file__, '..', 'mito_frontend.js'))) as f:
    js_code_from_file = f.read()
with open(os.path.normpath(os.path.join(__file__, '..', 'mito_frontend.css'))) as f:
    css_code_from_file = f.read()


def get_mito_backend(
        *args: Any,
        comm_target_id: str='',
        analysis_to_replay: Optional[str]=None, # This is the parameter that tracks the analysis that you want to replay (NOTE: requires a frontend to be replayed!)
    ) -> MitoBackend:

    # We pass in the dataframes directly to the widget
    mito_backend = MitoBackend(*args, analysis_to_replay=analysis_to_replay) 

    # We create a callback that runs when the comm is actually created on the frontend
    def on_comm_creation(comm: Comm, open_msg: Dict[str, Any]) -> None:
        @comm.on_msg
        def _recv(msg):
            # Register handler for any incoming messages
            mito_backend.receive_message(msg['content']['data'])
        
        # Save the comm in the mito widget, so we can use this .send function
        mito_backend.mito_comm = comm

        # Send data to the frontend on creation, so the frontend knows that we have
        # actually registered the comm on the backend
        comm.send({'echo': open_msg['content']['data']}) # type: ignore

    # Register the comm target - so the callback gets called
    ipython = get_ipython() # type: ignore
    if ipython:
        ipython.kernel.comm_manager.register_target(comm_target_id, on_comm_creation)

    return mito_backend

def get_current_kernel_id() -> str:
    """
    Return the url of the current jupyter notebook server.
    
    Adapted from: https://github.com/jupyter/notebook/issues/3156
    """
    # path/to/file/kernel-c13462b6-c6a2-4e17-b891-7f7847204df9.json
    connection_file = get_connection_file() # type: ignore
    # kernel-c13462b6-c6a2-4e17-b891-7f7847204df9.json
    file_name: str = os.path.basename(connection_file) 
    return file_name[len('kernel-'):-len('.json')] 


def get_mito_frontend_code(kernel_id: str, comm_target_id: str, div_id: str, mito_backend: MitoBackend) -> str:
    js_code = js_code_from_file.replace('REPLACE_THIS_WITH_DIV_ID', div_id)
    js_code = js_code.replace('REPLACE_THIS_WITH_KERNEL_ID', kernel_id)
    js_code = js_code.replace('REPLACE_THIS_WITH_COMM_TARGET_ID', comm_target_id)
    js_code = js_code.replace('REPLACE_THIS_WITH_CSS', css_code_from_file)
    # NOTE: we encode these as utf8 encoded byte arrays, so that we can avoid having to do complicated things with 
    # replacing \t, etc, which is required because JSON.parse limits what characters are valid in strings (bah humbug)
    def to_uint8_arr(string: str) -> List[int]:
        return np.frombuffer(string.encode("utf8"), dtype=np.uint8).tolist()
    
    js_code = js_code.replace('sheetDataBytes = new Uint8Array([]);', f'sheetDataBytes = new Uint8Array({to_uint8_arr(mito_backend.steps_manager.sheet_data_json)});')
    js_code = js_code.replace('analysisDataBytes = new Uint8Array([]);', f'analysisDataBytes = new Uint8Array({to_uint8_arr(mito_backend.steps_manager.analysis_data_json)});')
    js_code = js_code.replace('userProfileBytes = new Uint8Array([]);', f'userProfileBytes = new Uint8Array({to_uint8_arr(mito_backend.get_user_profile_json())});')

    return js_code


def sheet(
        *args: Any,
        analysis_to_replay: Optional[str]=None, # This is the parameter that tracks the analysis that you want to replay (NOTE: requires a frontend to be replayed!)
        view_df: bool=False, # We use this param to log if the mitosheet.sheet call is created from the df output button,
        # NOTE: if you add named variables to this function, make sure argument parsing on the front-end still
        # works by updating the getArgsFromCellContent function.
    ) -> None:
    """
    Renders a Mito sheet. If no arguments are passed, renders an empty sheet. Otherwise, renders
    any dataframes that are passed. Errors if any given arguments are not dataframes or paths to
    CSV files that can be read in as dataframes.

    If running this function just prints text that looks like `MitoWidget(...`, then you need to 
    install the JupyterLab extension manager by running:

    python -m pip install mitoinstaller
    python -m mitoinstaller install

    Run this command in the terminal where you installed Mito. It should take 1-2 minutes to complete.

    Then, restart your JupyterLab instance, and refresh your browser. Mito should now render.

    NOTE: if you have any issues with installation, please email jake@sagacollab.com
    """
    # We throw a custom error message if we're sure the user is in
    # vs code or google collab (these conditions are more secure than
    # the conditons for checking if we're in JLab or JNotebook).
    if is_in_vs_code() or is_in_google_colab():
        log('mitosheet_sheet_call_location_failed', failed=True)
        raise Exception("The mitosheet currently only works in JupyterLab.\n\nTo see instructions on getting Mitosheet running in JupyterLab, find install instructions here: https://docs.trymito.io/getting-started/installing-mito")

    # If the user.json does not exist, we create it. This ensures if the file is deleted in between
    # when the package is imported and mitosheet.sheet is called, the user still gets a user.json. 
    # We don't need to upgrade as creating the file will automatically use the most recent version
    if not os.path.exists(USER_JSON_PATH):
        try_create_user_json_file()

    try:

        # Every Mitosheet has a different comm target, so they each create
        # a different channel to communicate over
        comm_target_id = get_new_id()

        mito_backend = get_mito_backend(*args, comm_target_id=comm_target_id, analysis_to_replay=analysis_to_replay)

        # Log they have personal data in the tool if they passed a dataframe
        # that is not tutorial data or sample data from import docs
        if mito_backend.steps_manager.data_type_in_mito == DataTypeInMito.PERSONAL:
            log('used_personal_data') 

    except:
        log('mitosheet_sheet_call_failed', failed=True)
        raise


    # Then, we log that the call was successful, along with all of it's params
    log(
        'mitosheet_sheet_call',
        {
            # NOTE: analysis name is the UUID that mito saves the analysis under
            'steps_manager_analysis_name': mito_backend.steps_manager.analysis_name,
            'num_args': len(args),
            'num_str_args': len([arg for arg in args if isinstance(arg, str)]),
            'num_df_args': len([arg for arg in args if isinstance(arg, pd.DataFrame)]),
            'df_index_type': [str(type(arg.index)) for arg in args if isinstance(arg, pd.DataFrame)],
            'view_df': view_df
        }
    )

    div_id = get_new_id()
    kernel_id = get_current_kernel_id()

    js_code = get_mito_frontend_code(kernel_id, comm_target_id, div_id, mito_backend)

    display(HTML(f"""<div id={div_id} class="mito-container-container">
        <script>
            {js_code}
        </script>
    </div>""")) # type: ignore


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains tests to make sure that the mito analytics test is
performing correctly
"""

from mitosheet.telemetry.telemetry_utils import PRINT_LOGS


def test_not_printing_logs():
    assert PRINT_LOGS is False



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Dict, List, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import TAB


class ExportToFileCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, export_type: str, file_name: str, sheet_index_to_export_location: Dict[int, str]):
        super().__init__(prev_state, post_state)
        self.export_type = export_type
        self.file_name = file_name
        self.sheet_index_to_export_location = sheet_index_to_export_location

    def get_display_name(self) -> str:
        return 'Export To File'
    
    def get_description_comment(self) -> str:

        return f"Exports {len(self.sheet_index_to_export_location)} to file {self.file_name}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        if self.export_type == 'csv':
            return [
                f'{self.post_state.df_names[sheet_index]}.to_csv("{export_location}", index=False)'
                for sheet_index, export_location in self.sheet_index_to_export_location.items()
            ], []
        elif self.export_type == 'excel':
            # If there is only one sheet being exported, we can avoid creating the pd.ExcelWriter
            if len(self.sheet_index_to_export_location) == 1:
                for sheet_index, export_location in self.sheet_index_to_export_location.items():
                    return [f'{self.post_state.df_names[sheet_index]}.to_excel("{self.file_name}", sheet_name="{export_location}", index={False})'], []

            return [f'with pd.ExcelWriter("{self.file_name}") as writer:'] + [
                f'{TAB}{self.post_state.df_names[sheet_index]}.to_excel(writer, sheet_name="{export_location}", index={False})'
                for sheet_index, export_location in self.sheet_index_to_export_location.items()
            ], ['import pandas as pd']
        else:
            raise ValueError(f'Not a valid file type: {self.export_type}')